@@ -16,8 +16,9 @@ from xtlib import utils
 from xtlib import errors
 from xtlib import scriptor
 from xtlib import constants
-from xtlib import file_utils
 from xtlib import job_helper
+from xtlib import file_utils
+from xtlib import store_utils
 from xtlib.console import console
 from xtlib.helpers.xt_config import XTConfig
 from xtlib.report_builder import ReportBuilder
@@ -66,6 +67,7 @@ class AzureBatch(BackendBase):
         self.core = core
         self.config = config
         self.username = username
+        self.custom_image_name = None
 
         # first, ensure we have a config file
         if config:
@@ -90,6 +92,16 @@ class AzureBatch(BackendBase):
     def get_name(self):
         return "batch"
 
+    # API call
+    def provides_container_support(self):
+        '''
+        Returns:
+            returns True if docker run command is handled by the backend.
+        Description:
+            For Azure Batch, we use the XT-level container support. We had problems getting the Azure Batch container support to work.
+        '''
+        return False
+
     def adjust_run_commands(self, job_id, job_runs, using_hp, experiment, service_type, snapshot_dir, args):
         '''
         This method is called to allow the backend to inject needed shell commands before the user cmd.  At the
@@ -106,7 +118,9 @@ class AzureBatch(BackendBase):
             node_id = "node" + str(i)
             br = box_runs[0]
             box_info = br["box_info"]
+            run_name = br["run_name"]
             actions = ["data", "model"]
+            box_secret = br["box_secret"]
 
             is_windows = False
 
@@ -116,68 +130,95 @@ class AzureBatch(BackendBase):
             # just create the fn_wrapped for the first node (but use for all nodes)
             if i == 0:
                 node_cmd = " ".join(cmd_parts)
-                run_name = br["run_name"]
 
                 fn_wrapped = self.wrap_cmd_for_batch(node_cmd, node_id, store_data_dir, data_action, data_writable, 
-                    store_model_dir, model_action, model_writable, storage_name, storage_key, snapshot_dir, run_name, args)
+                    store_model_dir, model_action, model_writable, storage_name, storage_key, snapshot_dir, 
+                    box_secret, args)
 
-            # we pass the NODE_ID as a parameter to the fn_wrapped script
-            wrapped_parts = ['/bin/bash', '--login', os.path.basename(fn_wrapped) + " " + node_id]
+            # NOTE: we pass the NODE_INDEX and RUN_NAME as parameters to the fn_wrapped script
+            # so that we can share a single script among all nodes
+            wrapped_parts = ['/bin/bash', '--login', '{} {} {}'.format(os.path.basename(fn_wrapped), i, run_name)]
             run_specs["cmd_parts"] = wrapped_parts
 
+            merge_batch_logs = args["merge_batch_logs"]
+            if merge_batch_logs:
+                # recompose this command to pass along all script arguments
+                wrapped_parts = ['/bin/bash', '--login', '{} $*'.format(os.path.basename(fn_wrapped))]
+    
+                # create an STDBOTH.txt that merged stdout and stderr
+                node_cmd = "{} > ../stdboth.txt 2>&1".format(" ".join(wrapped_parts))
+
+                # create a helper file to call our node cmd
+                fn_helper = snapshot_dir + "/" + constants.FN_BATCH_HELPER
+                file_utils.write_text_file(fn_helper, node_cmd)
+
+                run_specs["cmd_parts"] = ["bash", constants.FN_BATCH_HELPER, str(i), run_name]
+
+                console.diag(run_specs)
+
     def wrap_cmd_for_batch(self, user_cmd, node_id, store_data_dir, data_action, data_writable, 
-        store_model_dir, model_action, model_writable, storage_name, storage_key, snapshot_dir, run_name, args):
+        store_model_dir, model_action, model_writable, storage_name, storage_key, snapshot_dir, box_secret, args):
 
         cmds = []
 
-        code_zip = args["code_zip"]
-        if code_zip and code_zip != "none":
-            # NOTE: usage of $ENVVAR must be surrounded by DOUBLE quotes
-            self.append(cmds, 'sudo apt install unzip')
-            self.append(cmds, 'unzip -n ' + constants.CODE_ZIP_FN + " > __unzip__.log")
+        gen_for_windows = False
+        self.capture_setup_cmds = args["capture_setup_cmds"]
 
-        super().add_first_cmds(cmds, "wrapped.sh", change_dir=False)
-        super().add_setup_cmds(cmds, pip_freeze=True, args=args)
-        super().add_xt_setup_cmds(cmds, is_windows=False)
-        super().add_report_cmds(cmds)
+        # default for backends is to pass these 2 args to their __wrapped__.xx script (philly is different)
+        node_id = "node$1"
+        run_name = "$2"
+        docker_cmd = args["docker_cmd"]
+        job_id = args["job_id"]
 
-        # we use pip install --local, so add it to the path
-        self.append(cmds, 'export PATH=$HOME/.local/bin:$PATH')
+        # for wrapper building functions to access 
+        self.node_id = node_id
+        self.run_name = run_name
+        
+        add_pre_cmds = False
 
-        # since node id is the only different between nodes (in this wrapped.sh file), make it a parameter
-        self.append(cmds, 'export XT_NODE_ID=$1')       #  + node_id)
+        if add_pre_cmds:
+            # add env vars (the Batch Task environment_setting property seems to be broken
+            env_vars = dict(args["env_vars"])
+            scriptor.add_controller_env_vars(env_vars, self.config, box_secret, None)
 
-        # emit cmds to MOUNT run store OUTPUT path
-        workspace = args["workspace"]
-        store_path = "runs/{}/output".format(run_name)
-        self.emit_mount_cmds(cmds, storage_name, storage_key, container=workspace, store_path=store_path, 
-            mnt_path="$HOME/mnt/workspace", is_writable=True, install_blobfuse=False, 
-            sudo_available=False, use_username=True, use_allow_other=False, env_var_name="XT_OUTPUT_DIR", 
-            env_var_name2="XT_OUTPUT_MNT")
+            self.append_title(cmds, "PRE-SETUP cmds:")
+            for name, value in env_vars.items():
+                cmd = "export {}={}".format(name, value)
+                self.append(cmds, cmd)
 
-        # emit cmds to MOUNT or DOWNLOAD data
-        self.process_action(cmds, data_action, "$HOME/mnt/data", "XT_DATA_DIR", "XT_DATA_MNT", utils.DATA_STORE_ROOT, 
-            store_data_dir, data_writable, storage_name, storage_key, sudo_available=False, use_allow_other=False)
+        # our initial working dir: .../wd
+        # this will change us to: .../wd/.xt/cwd
+        self.add_first_cmds(cmds, homebase="$HOME", cwd="$HOME/.xt/cwd", args=args)
 
-        # emit cmds to MOUNT or DOWNLOAD model
-        self.process_action(cmds, model_action, "$HOME/mnt/models", "XT_MODEL_DIR", "XT_MODEL_MNT", utils.MODELS_STORE_ROOT, 
-            store_model_dir, model_writable, storage_name, storage_key, sudo_available=False, use_allow_other=False)
+        self.append_title(cmds, "UNZIP code/bootstrap files:")
 
-        self.append(cmds, "echo final PIP FREEZE:")
-        self.append(cmds, "pip freeze > __final_pip_freeze__.log")
+        # unzip files from wd to current directory
+        if docker_cmd:
+            fn_bootstrap = "/usr/src/" + constants.CODE_ZIP_FN
+            self.append_unzip(cmds, fn_bootstrap, ".")
+        else:
+            fn_bootstrap = "../../" + constants.CODE_ZIP_FN
+            self.append_unzip(cmds, fn_bootstrap, ".")
+            self.append(cmds, "rm {}".format(fn_bootstrap))
 
-        # run user's cmd
-        self.append(cmds, user_cmd)
+        self.append_dir(cmds)
 
-        # remove empty cmds
-        cmds = [cmd for cmd in cmds if cmd]
+        super().add_setup_cmds(cmds, remove_zip=False, gen_for_windows=gen_for_windows, 
+            args=args)
 
-        fn_wrapped = snapshot_dir + "/" + constants.FN_WRAPPED_CMDS
-        scriptor.write_script_file(cmds, fn_wrapped, for_windows=False)
+        super().add_other_cmds(cmds, args)
+        super().add_report_cmds(cmds, args=args)
 
-        # copy to submit-logs
-        utils.copy_to_submit_logs(args, fn_wrapped)
+        # since node id is the only different between nodes (in this __wrapped__.sh file), make it a parameter
+        self.append(cmds, 'export XT_NODE_ID=node$1')       #  + node_id)
 
+        actions = ["data", "model"]
+
+        self.add_mount_cmds(cmds, False, storage_name, storage_key, 
+            actions, data_action, model_action, store_data_dir, store_model_dir, data_writable, model_writable,
+            use_username=False, use_allow_other=False, args=args)
+
+        fn_wrapped = self.create_wrapper_and_inner(cmds, user_cmd, args)
         return fn_wrapped
 
     def print_things(self, things, columns):
@@ -244,7 +285,7 @@ class AzureBatch(BackendBase):
         #console.print("pool_info=", pool_info)
 
         vm_size = compute_def["vm-size"]
-        vm_image = compute_def["azure-image"]
+        vm_image = utils.safe_value(compute_def, "azure-image")
         num_nodes = compute_def["nodes"]
         use_low_pri = compute_def["low-pri"]
 
@@ -264,6 +305,7 @@ class AzureBatch(BackendBase):
 
         ws_name = args["workspace"]
         is_distributed = args["distributed"]
+        merge_batch_logs = args["merge_batch_logs"]
 
         # first, build a list of runs for each node in our azure batch pool
         node_records = []
@@ -282,7 +324,10 @@ class AzureBatch(BackendBase):
                 fb.feedback("{}/{}".format(workspace, run_name))
 
             node_id = "node" + str(n)
-            service_info_by_node[node_id] = {"job_id": job_id, "node_id": node_id, "task_index": n}
+            batch_job_id = "{}_{}".format(ws_name, job_id)
+
+            service_info_by_node[node_id] = {"workspace": ws_name, "job_id": job_id, "batch_job_id": batch_job_id, 
+                "node_id": node_id, "task_index": n, "merge_batch_logs": merge_batch_logs}
 
         # "hold" is now used to hold open the xt controller task AND the created pool
         auto_pool = not args["hold"]
@@ -293,7 +338,7 @@ class AzureBatch(BackendBase):
 
         fb.feedback("job submitted", is_final=True)
 
-        service_run_info = {"batch_job_id": job_id, "pool_id": pool_id}
+        service_run_info = {"batch_job_id": self.batch_job_id, "pool_id": pool_id}
 
         return service_run_info, service_info_by_node    
 
@@ -312,7 +357,7 @@ class AzureBatch(BackendBase):
 
         # build RESOURCE FILES for each file we have uploaded into "before"  directory
         before_path = "before/code"      # job store
-        after_path = "node-{}/after".format(node_index)         # run store
+        after_path = "node-{}/after/service_logs".format(node_index)         # run store
 
         # use blob service from store 
         store_provider = self.store.helper.provider
@@ -320,10 +365,18 @@ class AzureBatch(BackendBase):
         # get list of BEFORE blobs (captured by runner, with controller adjustments)
         before_blob_path = file_utils.path_join(constants.JOBS_DIR, job_id, before_path)
 
+        workspace = args["workspace"]
+        jobs_container = store_utils.get_jobs_container(workspace)
+
+        fake_submit = utils.safe_value(args, "fake_submit")
+
         # BATCH BUG WORKAROUND: tried to create a resource file for FOLDER instead of enumerating all blobs
         # but always get "blob not found" error (before, before/, before/** - no form works)
         #blob_names = [before_blob_path + "/**"]
-        blob_names = store_provider.list_blobs(constants.INFO_CONTAINER, before_blob_path)
+        blob_names = store_provider.list_blobs(jobs_container, before_blob_path)
+
+        if not fake_submit:
+            assert len(blob_names) > 0
 
         # for the DEST FILENAME on the node, strip off the blob path prefix
         bbp_len = 1 + len(before_blob_path)    # +1 to remove the trailing "/"
@@ -334,13 +387,13 @@ class AzureBatch(BackendBase):
         console.diag("  node_file_names=" + str(node_file_names))
 
         # use our helper to convert blobs and filenames to resource files
-        node_res_files = self.convert_blobs_to_resource_files(constants.INFO_CONTAINER, blob_names, node_file_names)
+        node_res_files = self.convert_blobs_to_resource_files(jobs_container, blob_names, node_file_names)
 
         # convert OUTPUT WILDCARDS to output files
         output_file_list = ["controller.log", "../*.txt"]
 
         after_blob_path = constants.JOBS_DIR + "/" + job_id + "/" + after_path
-        node_output_files = self.build_output_files(constants.INFO_CONTAINER, after_blob_path, output_file_list)
+        node_output_files = self.build_output_files(workspace, after_blob_path, output_file_list)
 
         if static_run_cmds:
             fb.feedback("  adding node: {}, run(s): {}".format(box_name.upper(), run_names.upper()), is_final=True)
@@ -360,10 +413,12 @@ class AzureBatch(BackendBase):
         self.auto_pool = auto_pool
         self.description = description
 
-        #self.ws_name = ws_name
-        self.job_id = job_id
-        self.pool_id = "__pool" + job_helper.get_num_from_job_id(job_id) + "__"
-        #console.print("pool_id=", self.pool_id + ", job_id=", self.job_id)
+        # must now qualify job_id with ws_name to prevent duplicate job names
+        workspace = args["workspace"]
+        batch_job_id = workspace + "_" + job_id
+
+        self.batch_job_id = batch_job_id
+        self.pool_id = "_pool_{}_{}_".format(workspace, job_id)
 
         self.vm_size = vm_size
         self.azure_image = vm_image 
@@ -376,7 +431,7 @@ class AzureBatch(BackendBase):
 
         if not fake_submit:
             # create our pool and job together 
-            self.create_pool_and_job(is_distributed)
+            self.create_pool_and_job(is_distributed, args)
             
         # add the specified tasks (commands) to our job
         self.add_tasks_to_job(node_records, args)
@@ -439,9 +494,10 @@ class AzureBatch(BackendBase):
         job_id = service_node_info["job_id"]
         node_id = service_node_info["node_id"]
         node_index = utils.node_index(node_id)
+        workspace = service_node_info["workspace"]
 
         state, ip_addr, controller_port, tensorboard_port = \
-            self.get_azure_box_addr(job_id, node_index)
+            self.get_azure_box_addr(workspace, job_id, node_index)
 
         if not (ip_addr and controller_port):
             errors.service_error("Node not available (node state: {})".format(state))
@@ -451,7 +507,7 @@ class AzureBatch(BackendBase):
 
         return cs
 
-    def get_azure_box_addr(self, job_id, node_index):
+    def get_azure_box_addr(self, ws_name, job_id, node_index):
         ip_addr = None
         port = None
         state = None
@@ -463,8 +519,9 @@ class AzureBatch(BackendBase):
 
         # XT always has exactly 1 task running on each node (xt controller), so
         # we can reply on task[x] running on node index x
+        batch_job_name = ws_name + "_" + job_id
         task_id = "task" + str(node_index)
-        task = self.batch_client.task.get(job_id, task_id)
+        task = self.batch_client.task.get(batch_job_name, task_id)
 
         state = task.state
         if state in ["running", "completed"]:
@@ -525,7 +582,8 @@ class AzureBatch(BackendBase):
 
         return network_config
 
-    def create_pool_and_job(self, is_distributed):
+    def create_pool_and_job(self, is_distributed, args):
+        # get the required Azure VM image 
         props = self.config.get("azure-batch-images", self.azure_image)
         if not props:
             errors.config_error("No config file entry found in [azure-batch-images] section for azure-image=" + self.azure_image)
@@ -536,11 +594,35 @@ class AzureBatch(BackendBase):
         version = props["version"]
         node_agent_sku_id = props["node-agent-sku-id"]
 
-        # console.print("vm-image=", self.azure_image, ", props=", props)
-
         img_ref = batchmodels.ImageReference(publisher=publisher, offer=offer, sku=sku, version=version)
         vmc = batchmodels.VirtualMachineConfiguration(image_reference=img_ref, node_agent_sku_id=node_agent_sku_id)
 
+        # support for running a user-specified custom container
+        target = self.compute
+
+        # docker_name = args["docker"]
+        # docker_image, login_server, docker_registry = self.config.get_docker_info(target, docker_name, required=False)
+
+        # if docker_image:
+
+        #     # create/assign container_config
+        #     container_config = batch.models.ContainerConfiguration()
+        #     vmc.container_configuration = container_config
+
+        #     creds_required = utils.make_numeric_if_possible( utils.safe_value(docker_registry, "login") )
+        #     if creds_required:
+        #         # get username/password to log into private docker registry service (Azure)
+        #         username = utils.safe_value(docker_registry, "username")
+        #         password = utils.safe_value(docker_registry, "password")
+
+        #         # create/add registry
+        #         container_registry = batch.models.ContainerRegistry(registry_server=login_server, user_name=username, password=password)
+        #         container_config.container_registries = [container_registry]
+            
+        #     # create/add container_config
+        #     self.custom_image_name = login_server + "/" + docker_image
+        #     container_config.container_image_names = [self.custom_image_name]
+
         network_config = self.create_network_config()
 
         max_tasks_per_node = None
@@ -609,7 +691,7 @@ class AzureBatch(BackendBase):
             pool_info = batch.models.PoolInformation(pool_id=self.pool_id)  # , auto_pool_specification= auto_pool)
 
         # CREATE THE JOB (but doesn't launch it yet)
-        batch_job = batch.models.JobAddParameter(id=self.job_id, pool_info=pool_info, 
+        batch_job = batch.models.JobAddParameter(id=self.batch_job_id, pool_info=pool_info, 
             on_all_tasks_complete="terminateJob")
 
         self.batch_client.job.add(batch_job)
@@ -648,7 +730,6 @@ class AzureBatch(BackendBase):
         # we always use exactly 1 task per node (xt controller)
         tasks = []
         for idx, node_record in enumerate(node_records):
-            #cmd = "/bin/bash -c \"cat {}\"".format(input_file.file_path)
             node_cmd = node_record["node_cmd"]
             node_res_files = node_record["node_res_files"]
             node_output_files = node_record["node_output_files"]
@@ -670,8 +751,14 @@ class AzureBatch(BackendBase):
                 es = batchmodels.EnvironmentSetting(name=key, value=str(value))
                 env_var_list.append(es)
 
+            if self.custom_image_name:
+                tcs = batch.models.TaskContainerSettings(image_name=self.custom_image_name)
+            else:
+                tcs = None
+
             task_param = batch.models.TaskAddParameter(id=task_id, command_line=node_cmd, environment_settings=env_var_list,
-                resource_files=node_res_files, user_identity=elevated_user, output_files=node_output_files)
+                resource_files=node_res_files, user_identity=elevated_user, output_files=node_output_files, 
+                container_settings=tcs)
 
             tasks.append(task_param)
 
@@ -679,11 +766,11 @@ class AzureBatch(BackendBase):
 
         if not fake_submit:
             # this statement launches the job
-            self.batch_client.task.add_collection(self.job_id, tasks)
+            self.batch_client.task.add_collection(self.batch_job_id, tasks)
 
         # copy to submit-logs
         tasks_data = self.make_submit_data_serializable(tasks)
-        dd = {"job_id": self.job_id, "tasks": tasks_data}
+        dd = {"job_id": self.batch_job_id, "tasks": tasks_data}
         utils.copy_data_to_submit_logs(args, dd, "batch_submit.json")
 
         ### bug workaround: setting "on_all_tasks_complete" below doesn't seem to work so
@@ -840,7 +927,7 @@ class AzureBatch(BackendBase):
 
         return records
 
-    def cancel_job_node(self, store, job_id, node_index, run_datas):
+    def cancel_job_node(self, store, ws_name, job_id, node_index, run_datas):
         pool_id = None
         node_id = None
         task_killed = False
@@ -851,7 +938,8 @@ class AzureBatch(BackendBase):
         # terminate the TASK
         #console.print("canceling: job={}, node_index={}, run_names={}".format(job_id, node_index, full_run_names))
 
-        task = self.batch_client.task.get(job_id, "task" + str(node_index))
+        batch_job_name = ws_name + "_" + job_id
+        task = self.batch_client.task.get(batch_job_name, "task" + str(node_index))
         before_status = str(task.state)
         #console.print("task.state=", task.state)
 
@@ -874,7 +962,7 @@ class AzureBatch(BackendBase):
         # wrap-up each run (logging, capture)
         cancel_results = []
 
-        if task_killed and run_datas:
+        if run_datas:
             for run_data in run_datas:    
                 ws = run_data["ws_name"]
                 run_name = run_data["run_name"]
@@ -895,7 +983,7 @@ class AzureBatch(BackendBase):
 
     def get_job_status(self, job_id):
         if not job_id:
-            job_id = self.job_id
+            job_id = self.batch_job_id
 
         console.print("get_job_status (azure): job_id=", job_id)
 
@@ -917,7 +1005,7 @@ class AzureBatch(BackendBase):
         return status
 
     def attach_task_to_console(self, job_id, run_name):
-        self.job_id = job_id
+        self.batch_job_id = job_id
 
         self.create_batch_client()
 
@@ -931,7 +1019,7 @@ class AzureBatch(BackendBase):
             while True:   
                 
                 # get a dict of the stats for each task
-                task_counts = self.batch_client.job.get_task_counts(self.job_id)
+                task_counts = self.batch_client.job.get_task_counts(self.batch_job_id)
 
                 elapsed = utils.elapsed_time(start)
                 console.print("waiting for queued job to start... (elapsed time: {}).format(elapsed)", end="\r")
@@ -961,31 +1049,31 @@ class AzureBatch(BackendBase):
             console.print("\txt attach " + run_name)
         else:
             #----- stream output to console ----
-            self.print_task_output(self.job_id, 0)
+            self.print_task_output(self.batch_job_id, 0)
 
     def print_task_output(self, job_id, task_index):
 
         # get task_id
-        tasks = self.batch_client.task.list(self.job_id)
+        tasks = self.batch_client.task.list(self.batch_job_id)
         console.print("task.list len=", len(tasks))
         
         task = next(iter(tasks), None)
         if not task:
             console.print("error - job has no tasks")
         else:
-            task_info = self.batch_client.task.get(self.job_id, task.id)
+            task_info = self.batch_client.task.get(self.batch_job_id, task.id)
             node_info = task_info.node_info
 
             if node_info:
                 # node has been allocated and not yet released
-                stream = self.batch_client.file.get_from_task(self.job_id, task.id, "stdout.txt")
+                stream = self.batch_client.file.get_from_task(self.batch_job_id, task.id, "stdout.txt")
 
                 while True:
                     for data in stream:
                         text = data.decode("utf-8")
                         console.print(text)
 
-                    task_counts = self.batch_client.job.get_task_counts(self.job_id)
+                    task_counts = self.batch_client.job.get_task_counts(self.batch_job_id)
                     if task_counts.running == 0:
                         console.print("<task terminated>")
                         break
@@ -1002,17 +1090,17 @@ class AzureBatch(BackendBase):
         """
         console.print('Printing task output...')
 
-        tasks = self.batch_client.task.list(self.job_id)
+        tasks = self.batch_client.task.list(self.batch_job_id)
 
         for task in tasks:
-            console.print("getting output for job={}, task={}".format(self.job_id, task.id))
+            console.print("getting output for job={}, task={}".format(self.batch_job_id, task.id))
             
-            task_info = self.batch_client.task.get(self.job_id, task.id)
+            task_info = self.batch_client.task.get(self.batch_job_id, task.id)
             node_info = task_info.node_info
 
             if node_info:
                 node_id = node_info.node_id
-                stream = self.batch_client.file.get_from_task(self.job_id, task.id, "stdout.txt")
+                stream = self.batch_client.file.get_from_task(self.batch_job_id, task.id, "stdout.txt")
                 file_text = self._stream_to_text(stream)
 
                 console.print("\nTask: {}, Node: {}, Standard output:".format(task.id, node_id))
@@ -1080,10 +1168,10 @@ class AzureBatch(BackendBase):
             batch_client.pool.delete(self.pool_id)
             self.pool_id = None
 
-        if self.job_id:
+        if self.batch_job_id:
             console.print("deleting self...")
-            batch_client.self.delete(self.job_id)
-            self.job_id = None
+            batch_client.self.delete(self.batch_job_id)
+            self.batch_job_id = None
 
     def cancel_runs_by_names(self, workspace, run_names, box_name):
         '''
@@ -1106,7 +1194,8 @@ class AzureBatch(BackendBase):
         cancelled_by_controller = False
 
         # is the controller available?
-        azure_batch_state, ip_addr, controller_port, tensorboard_port = self.get_azure_box_addr(job_id, node_index)
+        azure_batch_state, ip_addr, controller_port, tensorboard_port = \
+            self.get_azure_box_addr(workspace, job_id, node_index)
 
         if ip_addr and controller_port:
             # add workspace to run_names
@@ -1140,7 +1229,7 @@ class AzureBatch(BackendBase):
         if not cancelled_by_controller:
             # run is in a node that waiting for a pool or has left pool; just kill the whole node
             run_datas = [ {"ws_name": workspace, "run_name": run_name} for run_name in run_names ]
-            cancel_results, pool_id = self.cancel_job_node(self.store, job_id, node_index, run_datas)
+            cancel_results, pool_id = self.cancel_job_node(self.store, workspace, job_id, node_index, run_datas)
 
         return cancel_results        
 
@@ -1191,7 +1280,7 @@ class AzureBatch(BackendBase):
 
         return cancel_results_by_box
 
-    def get_active_jobs(self):
+    def get_active_jobs(self, ws_name):
         ''' return a list of job_id's running on this instance of Azure Batch '''
         if not self.batch_client:
             self.create_batch_client()
@@ -1205,7 +1294,7 @@ class AzureBatch(BackendBase):
         return job_ids
 
 
-    def cancel_runs_by_user(self, box_name):
+    def cancel_runs_by_user(self, ws_name, box_name):
         '''
         Args:
             box_name: the name of the box the runs ran on (pool service)
@@ -1216,7 +1305,7 @@ class AzureBatch(BackendBase):
         cancel_results_by_box = {}
 
         # get list of active jobs from batch
-        job_ids = self.get_active_jobs()
+        job_ids = self.get_active_jobs(ws_name)
         console.diag("after get_active_jobs()")
 
         if job_ids:
@@ -1229,7 +1318,7 @@ class AzureBatch(BackendBase):
 
             fields_dict = {"runs_by_box": 1}
 
-            job_records = mongo.get_info_for_jobs(filter_dict, fields_dict)
+            job_records = mongo.get_info_for_jobs(ws_name, filter_dict, fields_dict)
             console.diag("after get_info_for_jobs()")
 
             # cancel each ACTIVE job
@@ -1245,13 +1334,14 @@ class AzureBatch(BackendBase):
 
     def get_node_task(self, service_node_info):
         # get task_id
-        self.create_batch_client()
-
-        job_id = service_node_info["job_id"]
+        if not self.batch_client:
+            self.create_batch_client()
+        
+        batch_job_id = service_node_info["batch_job_id"]
         task_index = service_node_info["task_index"]
 
         task_id = "task" + str(task_index)
-        task = self.batch_client.task.get(job_id, task_id)
+        task = self.batch_client.task.get(batch_job_id, task_id)
 
         return task
 
@@ -1261,28 +1351,146 @@ class AzureBatch(BackendBase):
         task = self.get_node_task(service_node_info)
         return task.state.value
 
-    # BACKEND API
-    def read_log_file(self, service_node_info, log_name, start_offset=0, end_offset=None, 
+   # BACKEND API
+    def log_file_iterator(self, service_node_info, log_name, start_offset=0, end_offset=None, 
         encoding='utf-8', use_best_log=True):
         '''
-        returns subset of specified log file as text.
+        returns an iterator for the specified log file.
         '''
         job_id = service_node_info["job_id"]
-        task = self.get_node_task(service_node_info)
+        merge_batch_logs = utils.safe_value(service_node_info, "merge_batch_logs")
+        node_id = service_node_info["node_id"]
+
+        new_text = None
+        batch_status = None
+        simple_status = None
+        got_text = False
 
         if log_name is None or log_name in ["primary", "stdout", "stdout.txt"]:
-            file_path = "stdout.txt"
+            file_path = "stdboth.txt" if merge_batch_logs else "stdout.txt"
         elif log_name in ["secondary", "error", "stderr", "stderr.txt"]:
             file_path = "stderr.txt"
         else:
             file_path = log_name
 
+        if start_offset == 0:
+            # try to first read from job storage (if task has completed)
+            node_index = utils.node_index(node_id)
+            job_path = "node-{}/after/service_logs/{}".format(node_index, file_path)
+            if self.store.does_job_file_exist(job_id, job_path):
+                console.print("==> reading log from job storage...")
+
+                new_text = self.store.read_job_file(job_id, job_path)
+                batch_status = "completed"
+                simple_status = "completed"
+                got_text = True
+
+                result = {"new_text": new_text, "simple_status": simple_status, "log_name": log_name, 
+                    "service_status": batch_status, "file_path": file_path}
+
+                # here we just yield a single result
+                yield result                
+
+        if not got_text:
+            # this call can take up to 15 secs
+            task_index = service_node_info["task_index"]
+            task_id = "task" + str(task_index)
+            task = None     # delay creation until needed
+
+            if not self.batch_client:
+                self.create_batch_client()
+
+            next_offset = start_offset
+            if not end_offset:
+                end_offset = 1024*1024*1024*16     # 16 GB should be big enough for a log file
+
+            while True:
+                try:
+                    gft_opts = batchmodels.FileGetFromTaskOptions(ocp_range='bytes={}-{}'.format(next_offset, end_offset))
+
+                    # read currently available data
+                    stream = self.batch_client.file.get_from_task(job_id, task_id, file_path=file_path, 
+                        file_get_from_task_options=gft_opts)
+
+                    output = io.BytesIO()
+                    for data in stream:
+                        output.write(data)
+
+                    new_bytes = output.getvalue()
+                    new_text = new_bytes.decode(encoding)
+                    next_offset += len(new_bytes)
+
+                    # if we are streaming output, assume we are running
+                    batch_status == "running"
+                    simple_status == "running"
+
+                except (batchmodels.BatchErrorException) as ex:
+                    # interpret this error as task has terminated
+                    console.diag("Exception reading batch log file: {}".format(ex))
+                    next_text = ""
+
+                if not new_text:
+                    # is task done?  update status 
+                    # CAUTION: this is a 15-30 second call
+                    if task is None:
+                        console.print("==> creating azure batch client...")
+                    task = self.get_node_task(service_node_info)
+                    batch_status = task.state.value
+                    simple_status = self.get_simple_status(batch_status)
+
+                result = {"new_text": new_text, "simple_status": simple_status, "log_name": log_name, 
+                    "service_status": batch_status, "file_path": file_path}
+
+                # we never stop yielding; its up to caller to examine status and stop iteration
+                yield result
+
+    # BACKEND API
+    def read_log_file(self, service_node_info, log_name, start_offset=0, end_offset=None, 
+        encoding='utf-8', use_best_log=True, service_context=None):
+        '''
+        returns subset of specified log file as text.
+        '''
+        job_id = service_node_info["job_id"]
+        batch_job_id = service_node_info["batch_job_id"]
+        node_id = service_node_info["node_id"]
+        workspace = service_node_info["workspace"]
+
+        # allow task to be cached in service_context (for multiple calls)
+        task = utils.safe_value(service_context, "task")
+
         new_text = None
         batch_status = None
         simple_status = None
         next_offset = 0
+        found_file = False
+
+        if log_name is None:
+            log_name = "stdout.txt"
+        file_path = os.path.basename(log_name)
+
+        # try to first read from job storage (if task has completed)
+        node_index = utils.node_index(node_id)
+
+        job_path = "node-{}/after/service_logs/{}".format(node_index, file_path)
+        if self.store.does_job_file_exist(workspace, job_id, job_path):
+            new_text = self.store.read_job_file(workspace, job_id, job_path)
+            batch_status = "completed"
+            simple_status = "completed"
+            found_file = True
+
+        job_path = "node-{}/after/xt_logs/{}".format(node_index, file_path)
+        if self.store.does_job_file_exist(workspace, job_id, job_path):
+            new_text = self.store.read_job_file(workspace, job_id, job_path)
+            batch_status = "completed"
+            simple_status = "completed"
+            found_file = True
+
+        if not found_file:
+
+            # this call can take up to 15 secs
+            if not task:
+                task = self.get_node_task(service_node_info)
 
-        if task:
             node_info = task.node_info
             batch_status = task.state.value
             simple_status = self.get_simple_status(batch_status)
@@ -1293,11 +1501,12 @@ class AzureBatch(BackendBase):
             gft_opts = batchmodels.FileGetFromTaskOptions(ocp_range='bytes={}-{}'.format(start_offset, end_offset))
 
             try:
-                stream = self.batch_client.file.get_from_task(job_id, task.id, file_path=file_path, 
+                stream = self.batch_client.file.get_from_task(batch_job_id, task.id, file_path=log_name, 
                     file_get_from_task_options=gft_opts)
 
                 # put streamed output into "output"
                 output = io.BytesIO()
+
                 for data in stream:
                     output.write(data)
 
@@ -1307,10 +1516,10 @@ class AzureBatch(BackendBase):
 
             except (batchmodels.BatchErrorException) as ex:
                 # interpret this error as task has terminated
-                pass
+                console.diag("Exception reading batch log file: {}".format(ex))
 
         return {"new_text": new_text, "simple_status": simple_status, "log_name": log_name, "next_offset": next_offset, 
-            "service_status": batch_status}
+            "service_status": batch_status, "file_path": file_path, "service_context": {"task": task}, "found_file": found_file}
 
     # API call
     def get_simple_status(self, status):
@@ -1351,8 +1560,8 @@ class AzureBatch(BackendBase):
 
         if simple_status != "completed":
             
-            job_id = service_node_info["job_id"]
-            self.batch_client.task.terminate(job_id, task.id)
+            batch_job_id = service_node_info["batch_job_id"]
+            self.batch_client.task.terminate(batch_job_id, task.id)
 
             # refresh the task for updated status
             task = self.get_node_task(service_node_info)
@@ -1363,3 +1572,9 @@ class AzureBatch(BackendBase):
 
         result = {"cancelled": cancelled, "service_status": service_status, "simple_status": simple_status}
         return result
+
+    # API call
+    def add_service_log_copy_cmds(self, cmds, dest_dir, store_path, args):
+        # this is done by the batch service at the end of our batch task
+        pass
+
