@@ -8,6 +8,7 @@ import sys
 import json
 import time
 import datetime
+import tempfile
 import numpy as np 
 
 from xtlib.client import Client
@@ -95,7 +96,7 @@ class ImplCompute(ImplBase):
     @flag("active", help="when specified, only active runs are reported")
     @flag("completed", help="when specified, only completed runs are reported")
     @option("escape-secs", type=int, help="how many seconds to wait before terminating the monitor loop")
-    @option("target", default="$xt-services.target", help="the name of the compute target to query for status")
+    @option("target", default=None, type=str, help="the name of the compute target to query for status")
     @option("username", default="$general.username", type=str, help="the username used to filter runs")
     @option("workspace", default="$general.workspace", type=str, help="the workspace used to filter runs")
     @option(name="job", help="query all boxes defined for the specified job")
@@ -111,6 +112,10 @@ class ImplCompute(ImplBase):
         The 'mirror' flag is used to return information about Grok server mirroring processes.
         '''
 
+        if not target:
+            ss_info = self.config.get_service_set_info()
+            target = utils.safe_value(ss_info, "target")
+
         if tensorboard:
             run_name = "tensorboard"
         elif mirror:
@@ -165,33 +170,82 @@ class ImplCompute(ImplBase):
         if jupyter:
             return self.monitor_with_jupyter(workspace, name)
 
-        if name.startswith("job"):
-            return self.monitor_job_node(name, jupyter, sleep, node_index, log_name, escape) 
+        if job_helper.is_job_id(name):
+            return self.monitor_job_node(workspace, name, sleep, node_index, log_name, escape) 
 
-        if name.startswith("run"):
+        if run_helper.is_run_name(name):
             rr = run_helper.get_run_record(self.store, workspace, name)
             job_id = rr["job_id"]
             node_index = rr["node_index"]
-            return self.monitor_job_node(job_id, jupyter, sleep, node_index, log_name, escape) 
+            return self.monitor_job_node(workspace, job_id, sleep, node_index, log_name, escape) 
 
         errors.syntax_error("name must be a job or run name: {}".format(name))
 
-    def monitor_job_node(self, job_id, jupyter, sleep, node_index, log_name, escape):
+    def show_monitor_help(self):
+        print("\ntraining interaction menu:")
+        print("  ESCAPE: stop monitoring job")
+        print("  control-c: cancel job")
+        print("  +: change to next node index")
+        print("  -: change to prev node index")
+        print()
+        print("  q: show queued jobs (for pool nodes)")
+        print("  t: show a prediction from TRAIN data")
+        print("  e: show a prediction from EVAL data")
+        print("  h: help (show this menu)")
+        print("  ?: show hyperparameter values")
+        print("  =: change the value of a hyperparameter")
+        print()
+
+    def send_user_request(self, request, ws_name, job_id, node_index):
+        rd = None
+
+        if request == "help":
+            self.show_monitor_help()
+        elif request == "set_hparam":
+            console.print()
+            name = input("enter HP name to set: ")
+            if name:
+                value = input("enter new value of {}: ".format(name))
+                rd = {"set_hparam": {"name": name, "value": value}}
+            console.print()
+        else:
+            rd = {request: 1}
+
+        if rd:
+            rd_text = json.dumps(rd)
+
+            # get the connection string for the job/node
+            cs_plus = job_helper.get_client_cs(self.core, ws_name, job_id, node_index)
+            cs = cs_plus["cs"]
+            box_secret = cs_plus["box_secret"]  
+
+            with XTClient(self.config, cs, box_secret) as xtc:
+                if xtc.connect():
+                    result = xtc.queue_user_request(rd_text)        
+
+    def monitor_job_node(self, workspace, job_id, sleep, node_index, log_name, escape):
 
         if node_index is None:
             node_index = 0
 
-        backend, job_info = job_helper.get_job_backend(self.store, self.core, job_id)
+        backend, job_info = job_helper.get_job_info_with_backend(self.store, self.core, workspace=workspace, 
+            job_id=job_id, node_index=node_index)
 
         node_id = utils.node_id(node_index)
         service_info_by_node = job_info["service_info_by_node"]  
         service_node_info = service_info_by_node[node_id]
+        pool_info = job_info["pool_info"]
+        node_count = len(service_info_by_node)
 
-        service_name = backend.get_name()
-        node_count = len(job_info["service_info_by_node"])
+        target = utils.safe_value(pool_info, "name")
+        merge_batch_logs = utils.safe_value(service_node_info, "merge_batch_logs")
+        if merge_batch_logs and log_name is None:
+            log_name = "stdboth.txt"
 
-        console.print("==> monitoring: {}, node{} [{}] (press escape or control-c to exit, +/- to change node)" \
-            .format(job_id, node_index, service_name), flush=True)
+        target_str = self.config.get_target_desc(target, backend)
+
+        console.print("xt monitor: press escape or control-c to exit, h for help menu")
+        console.print("==> monitoring: {}, node{} [{}]".format(job_id, node_index, target_str), flush=True)
 
         # the monitoring loop
         start_offset = 0
@@ -207,13 +261,16 @@ class ImplCompute(ImplBase):
         first_text_of_stream = True
         display_count = 0
         started = time.time()
+        last_status_shown = None
+        request_dict = {"t": "predict_train", "e": "predict_eval", "?": "show_hparams", "=": "set_hparam", "h": "help"}
 
         try:
             with KeyPressChecker() as checker:
 
                 while simple_status != "completed":
                     ch = checker.getch_nowait()
-                    #print(ch, end=" ")
+                    if ch is not None:
+                        console.print("detected user input key: {}".format(ch))
 
                     if ch == constants.ESCAPE or ch == constants.CONTROL_C:
                         break
@@ -228,7 +285,7 @@ class ImplCompute(ImplBase):
                         node_id = utils.node_id(node_index)
 
                         # set new context
-                        service_node_info = service_info_by_node[node_id]
+                        service_node_info = service_info_by_node[node_index]
                         start_offset = offset_by_node[node_id] if node_id in offset_by_node else 0
                         first_text_of_stream = not start_offset
                         console.print("==> switching to: node{}".format(node_index))
@@ -248,6 +305,11 @@ class ImplCompute(ImplBase):
                             else:
                                 console.print("  <no entries>")
 
+                    # turn this off for now (to easy to type in monitor window accidently and get interaction errors)
+                    # if ch in request_dict:
+                    #     request = request_dict[ch]
+                    #     self.send_user_request(request, workspace, job_id, node_index)
+
                     if sleep_count == sleeps_per_call:
 
                         # time to read the log file
@@ -258,16 +320,21 @@ class ImplCompute(ImplBase):
                         new_service = result["service_status"]
                         start_offset = result["next_offset"]
                         new_log = result["log_name"]
+                        file_path = utils.safe_value(result, "file_path")
 
-                        if new_log != log_name:
+                        if first_call or new_log != log_name:
                             log_name = new_log                
-                            console.print("==> node{} streaming log: {}".format(node_index, log_name))
+                            if file_path:
+                                console.print("==> node{} streaming log: {} ({})".format(node_index, log_name, file_path))
+                            else:
+                                console.print("==> node{} streaming log: {}".format(node_index, log_name))
                             first_text_of_stream = True
 
                         if new_service != service_status:
                             service_status = new_service
                             simple_status = new_simple
                             console.print("==> node{} status: {} ({})".format(node_index, service_status, simple_status))
+                            last_status_shown = simple_status
 
                         if new_text:
                             if node_count > 1:
@@ -306,12 +373,14 @@ class ImplCompute(ImplBase):
             console.print("==> monitoring cancelled (control-c detected)")
             ch = single_char_input("do you want to cancel the job? (y/n): ")
             if ch == "y":
-                self.cancel_job(job_id)
+                self.cancel_job(job_id, workspace)
                 show_final_status = False
 
-        if show_final_status:
+        if show_final_status and last_status_shown != simple_status:
             console.print("==> node{} status: {} ({})".format(node_index, service_status, simple_status))
 
+        console.print("end of monitoring: {}".format(job_id))
+
     def monitor_with_jupyter(self, workspace, run_name):
         if not self.is_aml_ws(workspace):
             errors.combo_error("the monitor command is only supported for Azure ML runs")
@@ -422,9 +491,11 @@ class ImplCompute(ImplBase):
         if not full_cmd.startswith("xt "):
             errors.syntax_error("command must start with 'xt ': {}".format(full_cmd))
 
+        # this temp dir cannot be removed immediately after job is submitted (TBD why)
         tmp_dir = file_utils.make_tmp_dir("rerun_cmd")
         job_id = self.store.get_job_id_of_run(workspace, run_name)
-        capture.download_before_files(self.store, job_id, workspace, run_name, tmp_dir, silent=True)
+        capture.download_before_files(self.store, job_id, workspace, run_name, tmp_dir, 
+            silent=True, log_events=False)
          
         # move to tmp_dir so files get captured correctly
         prev_cwd = os.getcwd()
@@ -500,12 +571,14 @@ class ImplCompute(ImplBase):
     def print_cancel_all_results(self, cancel_results_by_boxes):
         for target, results in cancel_results_by_boxes.items():
             console.print("Target: {}".format(target))
+
             for result in results:
-                console.print(
-                    "canceled: {}, service_status: {}, simple_status: {}".format(
-                        result.get("cancelled"),
-                        result.get("service_status"),
-                        result.get("simple_status")))
+                if result:
+                    console.print(
+                        "cancelled: {}, service_status: {}, simple_status: {}".format(
+                            result.get("cancelled"),
+                            result.get("service_status"),
+                            result.get("simple_status")))
 
     def get_runs_by_box(self, run_names, workspace=None):
         run_names, actual_ws = run_helper.parse_run_list(self.store, workspace, run_names)
@@ -544,7 +617,7 @@ class ImplCompute(ImplBase):
         for box_name, runs in runs_by_box.items():
             first_run = runs[0]
 
-            ws_name = first_run["ws"]
+            ws_name = first_run["ws_name"]
             # = [ws_name + "/" + run["run_name"] for run in runs]
             run_names = [run["run_name"] for run in runs]
             compute = first_run["compute"] if "compute" in first_run else "pool"    # some legacy support
@@ -558,13 +631,14 @@ class ImplCompute(ImplBase):
 
     #---- CANCEL JOB command ----
     @argument("job-id", help="the name of the job to cancel")
+    @option("workspace", default="$general.workspace", help="the workspace that contains the job to be cancelled")
     @example(task="cancel all runs in job3125", text="xt cancel job job3125")
     @command(kwgroup="cancel", kwhelp="cancels active runs, specified by name, job, or target", help="cancels the specified job and its active or queued runs")
-    def cancel_job(self, job_id):
+    def cancel_job(self, job_id, workspace):
 
         console.print("cancelling {}:".format(job_id))
 
-        service_job_info, service_info_by_node, backend = job_helper.get_service_job_info(self.store, self.core, job_id)
+        service_job_info, service_info_by_node, backend = job_helper.get_service_job_info(self.store, self.core, workspace, job_id)
         result_by_node = backend.cancel_job(service_job_info, service_info_by_node)
 
         for node_id, result in result_by_node.items():
@@ -574,7 +648,7 @@ class ImplCompute(ImplBase):
         wrapup_count = 0
 
         for node_id, result in result_by_node.items():
-            runs = run_helper.wrapup_runs_by_node(self.store, job_id, node_id, result)
+            runs = run_helper.wrapup_runs_by_node(self.store, workspace, job_id, node_id, result)
             wrapup_count += len(runs)
 
         console.print("  runs wrapped-up: {}".format(wrapup_count))
@@ -582,11 +656,12 @@ class ImplCompute(ImplBase):
     #---- CANCEL ALL command ----
     @argument("target", help="the name of the compute target whose runs will be cancelled")
     @hidden("username", default="$general.username", help="the username to log as the author of this run")
+    @option("workspace", default="$general.workspace", help="the workspace to cancel runs in")
     @example(task="cancel all runs for current user on Philly", text="xt cancel all philly")
     @command(kwgroup="cancel", kwhelp="cancels active runs, specified by name, job, or target", help="cancels all queued/active runs on the specified compute target")
-    def cancel_all(self, target, username):
+    def cancel_all(self, target, username, workspace):
 
-        backend = self.core.create_backend(target, username)
+        backend = self.core.create_backend(target, username=username)
         cancel_results_by_box = {}
 
         compute_info = self.config.get_compute_def(target)
@@ -595,11 +670,11 @@ class ImplCompute(ImplBase):
         if service_name == "pool":
             boxes = compute_info["boxes"]
             for box_name in boxes:
-                cancel_results = backend.cancel_runs_by_user(box_name)
+                cancel_results = backend.cancel_runs_by_user(ws_name=workspace, box_name=box_name)
             
                 cancel_results_by_box[box_name] = cancel_results
         else:
-            cancel_results = backend.cancel_runs_by_user(None)
+            cancel_results = backend.cancel_runs_by_user(ws_name=workspace, box_name=None)
             cancel_results_by_box[target] = cancel_results
 
         self.print_cancel_all_results(cancel_results_by_box)
@@ -712,7 +787,7 @@ class ImplCompute(ImplBase):
     @hidden("after-dirs", default="$after-files.after-dirs", help="the files and directories to upload after the run completes")
     @hidden("after-omit", default="$after-files.after-omit", help="the files and directories to omit from after uploading")
     @flag("after-upload", default="$after-files.after-upload", help="when true, the after files are upload when the run completes")
-    @hidden("option-prefix", default="$hyperparameter-search.option-prefix", help="the prefix to be used for specifying hyperparameter options to the script")
+    @hidden("capture_setup_cmds", type=bool, default="$logging.capture-setup-cmds", help="specifies if some setup cmds should output to a log file")
     @option("cluster", help="the name of the Philly cluster to be used")
     @hidden("code-dirs", default="$code.code-dirs", help="paths to the main code directory and dependent directories")
     @hidden("code-omit", default="$code.code-omit", help="the list wildcard patterns to omit uploading from the code files")
@@ -730,7 +805,7 @@ class ImplCompute(ImplBase):
     @flag("direct-run", default="$general.direct-run", help="when True, the script is run without the controller (on Philly or AML)")
     @flag("distributed", default="$general.distributed", help="when True, the multiple nodes will be put into distributed training mode")
     @hidden("distributed-training", default="$aml-options.distributed-training", help="the name of the backend process to use for distributed training")
-    @option("docker", help="the docker entry to use with the target (from one of the entries in the config file [dockers] section")
+    @option("docker", type=str, help="the name of the docker entry to use with the target (from one of the entries in the config file [dockers] section")
     @flag("dry-run", help="when True, the planned runs are displayed but not submitted")
     @hidden("early-policy", default="$early-stopping.early-policy", help="the name of the early-stopping policy to use (bandit, median, truncation, none)")
     @option("escape", type=int, default=0, help="breaks out of attach or --monitor loop after specified # of seconds")
@@ -745,11 +820,13 @@ class ImplCompute(ImplBase):
     @flag("hold", help="when True, the Azure Pool (VM's) are held open for debugging)")
     @option("hp-config", default="$hyperparameter-search.hp-config", type=str, help="the path of the hyperparameter config file")
     @hidden("log", default="$logging.log", help="specifes if run-related events should be logged")
+    @hidden("log-reports", default="$logging.log-reports", help="specifies if CPU, GPU, and MEMORY reports should be included in setup log")
     @option("low-pri", type=bool, help="when true, use low-priority (preemptable) nodes for this job")
     @option("max-minutes", default="$hyperparameter-search.max-minutes", type=int, help="the maximum number of minutes the run can execute before being terminated")
     @hidden("max-seconds", type=int, default="$aml-options.max-seconds", help="the maximum number of seconds this run will execute before being terminated")
     @option("max-runs", default="$hyperparameter-search.max-runs", type=int, help="the total number of runs across all nodes (for hyperparameter searches)")
     @hidden("maximize-metric", default="$general.maximize-metric", help="whether to minimize or maximize value of primary metric")
+    @hidden("merge-batch-logs", type=int, default="$logging.merge-batch-logs", help="replaces STDOUT.txt and STDERR.txt with STDBOTH.txt")
     @hidden("mirror-dest", default="$logging.mirror-dest", help="the location where mirrored information is store (none, storage, grok)")
     @hidden("mirror-files", default="$logging.mirror-files", help="the wildcard path that specifies files to be mirrored")
     @option("model-action", default="$model.model-action", help="the model action to take on the target, before run is started")
@@ -760,11 +837,11 @@ class ImplCompute(ImplBase):
     @option("monitor", default="$general.monitor", values=["new", "same", "none"], help="how to monitor primary run of the new job")
     @flag("multi-commands"   , help="the script file contains multiple run commands (one per line)")
     @option("nodes", type=int, help="the number of normal (non-preemptable) nodes to allocte for this run")
+    @hidden("option-prefix", default="$hyperparameter-search.option-prefix", help="the prefix to be used for specifying hyperparameter options to the script")
     @option("parent-script", type=str, help="path of script used to initialize the target for repeated runs of primary script")
-    @hidden("pip-freeze", default="$internal.pip-freeze", help="when true, installed pip packges are included in the setup log")
+    @hidden("pip-freeze", default="$logging.pip-freeze", help="when true, installed pip packges are included in the setup log")
     @hidden("primary-metric", default="$general.primary-metric", help="the name of the metric to use for hyperparameter searching")
     @option("queue", type=str, help="the name of the Philly queue to use when submitting this job")
-    #@option("repeat", type=int, help="the number of times to repeat the run, on each node")
     @hidden("remote-control", default="$general.remote-control", help="specifies if XT controller will listen for XT client commands")
     @hidden("report-rollup", default="$run-reports.report-rollup", help="whether to rollup metrics by primary metric or just use last reported metric set")
     @option("resume-name", help="when resuming a run, this names the previous run")
@@ -772,16 +849,18 @@ class ImplCompute(ImplBase):
     @option("schedule", default="static", values=["static", "dynamic"], help="specifies if runs are pre-assigned to each node or allocate on demand")
     @option("search-type", values=["random", "grid", "bayesian", "dgd"], default="$hyperparameter-search.search-type", help="the type of hyperparameter search to perform")
     @option("seed", default=None, help="the random number seed that can be used for reproducible HP searches")
+    @option("setup", type=str, help="the name of the setup entry to use with the target (from one of the entries in the config file [setups] section")
     @option("sku", help="the name of the Philly SKU to be used (e.g, 'G1')")
     @hidden("slack-factor", default="$early-stopping.slack-factor", type=float, help="(bandit only) specified as a ratio, the delta between this eval and the best performing eval")
     @hidden("slack-amount", default="$early-stopping.slack-amount", type=float, help="(bandit only) specified as an amount, the delta between this eval and the best performing eval")
-    @hidden("storage", default="$xt-services.storage", help="name of storage service to be used for this run")
+    @flag("static-search",  default="$hyperparameter-search.static-search", help="if true, hyperparameter searches are performed at job submit time, when possible")
+    @hidden("storage", default=None, type=str, help="name of storage service to be used for this run")
     @option("submit-logs", default=None, help="specifies a directory to which log files for the submit are saved")
-    @option("target", default="$xt-services.target", help="one of the user-defined compute targets on which to run")
+    @option("target", default=None, type=str, help="one of the user-defined compute targets on which to run")
     @hidden("truncation-percentage", default="$early-stopping.truncation-percentage", type=float, help="(truncation only) percent of runs to cancel at each eval interval")
     @option("use-gpu", type=bool, default="$aml-options.use-gpu", help="when True, the gpu(s) on the nodes will be used by the run")
     @option("username", default="$general.username", help="the username to log as the author of this run")
-    @hidden("user-managed", type=bool, default="$aml-options.user-managed", help="if true, it implies that the local machine or VM will be managed by the user")
+    # @hidden("user-managed", type=bool, default="$aml-options.user-managed", help="if true, it implies that the local machine or VM will be managed by the user")
     @option("vc", help="the name of the Philly virtual cluster to be used")
     @option("vm-size", help="the type of Azure VM computer to run on")
     @hidden("working-dir", default="$code.working-dir", type=str, help="the run's working directory, relative to the code directory")
@@ -804,23 +883,38 @@ class ImplCompute(ImplBase):
         '''
         #console.diag("run_cmd")
 
+        target = args["target"]
+        storage = args["storage"]
+
+        if not target:
+            ss_info = self.config.get_service_set_info()
+            target = utils.safe_value(ss_info, "target")
+            args["target"] = target
+
+        if not storage:
+            ss_info = self.config.get_service_set_info()
+            storage = utils.safe_value(ss_info, "storage")
+            args["storage"] = storage
+
         # add xt cmd to args
         cmd = get_dispatch_cmd()
         args["xt_cmd"] = "xt " + cmd if cmd else ""
 
-        runner = Runner(self.config, self.core)
+        with tempfile.TemporaryDirectory(prefix="import-") as temp_dir:
+            runner = Runner(self.config, self.core, temp_dir)
 
-        cmds, run_specs, using_hp, using_aml_hparam, sweeps_text, pool_info, job_id = \
-            runner.process_run_command(args)
+            cmds, run_specs, using_hp, using_aml_hparam, sweeps_text, pool_info, job_id = \
+                runner.process_run_command(args)
 
         monitor = args["monitor"]
         escape = args["escape"]
+        workspace = args["workspace"]
 
         if monitor == "same":
-            workspace = args["workspace"]
             self.monitor(name=job_id, workspace=workspace, escape=escape)
 
-        elif monitor == "new":
-            cmd = "xt --echo monitor {}".format(job_id)
-            process_utils.run_cmd_in_new_console(cmd)
+        elif monitor in ["bg", "new"]:
+            cmd = "xt --echo monitor {} --workspace={}".format(job_id, workspace)
+            noactivate = (monitor=="bg")
+            process_utils.run_cmd_in_new_console(cmd, noactivate=noactivate)
 
