@@ -5,6 +5,7 @@
 # run_helper.py: functions needed for processing run-related information
 import json
 import fnmatch
+from datetime import timedelta 
 
 from xtlib import qfe
 from xtlib import utils
@@ -14,8 +15,6 @@ from xtlib import file_utils
 from xtlib import job_helper
 
 from xtlib.console import console
-from xtlib.report_builder import ReportBuilder   
-from xtlib.storage import fixup_mongo_runs
 
 def expand_run_list(store, mongo, workspace, name_list):
     '''
@@ -43,9 +42,9 @@ def expand_run_list(store, mongo, workspace, name_list):
 
     if name_list:
         for entry in name_list:
-            if entry.startswith("run"):
+            if is_run_name(entry):
                 actual_ws = expand_run_entry(store, mongo, workspace, run_list, entry)
-            elif entry.startswith("job"):
+            elif job_helper.is_job_id(entry):
                 actual_ws = expand_job_entry(store, mongo, workspace, run_list, entry)
             else:
                 actual_ws = expand_experiment_name(store, mongo, workspace, run_list, entry)
@@ -72,16 +71,27 @@ def expand_run_entry(store, mongo, workspace, run_list, name):
     run_list += sub_list
     return actual_ws
 
-def expand_job_entry(store, mongo, workspace, run_list, name_entry):
+def expand_job_entry(store, mongo, workspace, run_list, job_term):
+    '''
+    Arguments:
+        store: an instance of Store (access to storage) 
+        mongo: an instance of Mongo (access to related mongoDB)
+        workspace: name of the workspace to search within
+        run_list: a list of run names being accumulated
+        job_term: a single job_id, or range of job_id's (job3-job5)
+    Processing:
+        Expand job_term to a list of job_id's and then for each job, 
+        add its run names to run_list
+    '''
     from xtlib import job_helper
 
-    # expand name_entry into a list of job names
-    job_list, actual_ws = job_helper.expand_job_list(store, mongo, workspace, [name_entry], can_mix=False)
+    # expand job name_entry into a list of job names
+    job_list, actual_ws = job_helper.expand_job_list(store, mongo, workspace, [job_term], can_mix=False)
 
     run_filter = {"job_id": {"$in": job_list}}
-    result = mongo.get_info_for_runs(workspace, run_filter, {"_id": 1})
-    if result:
-        names = [run["_id"] for run in result]
+    records = mongo.get_info_for_runs(workspace, run_filter, {"_id": 1})
+    if records:
+        names = [run["_id"] for run in records]
         run_list += names
 
     return actual_ws
@@ -123,14 +133,14 @@ def set_run_tags(store, mongo, name_list, tag_list, workspace, fd, clear):
     else:
         console.print("no matching runs found")
 
-def list_run_tags(store, mongo, name_list, tag_list, workspace):
-    run_list, actual_ws = expand_run_list(store, mongo, workspace, name_list)
+def list_run_tags(store, mongo, ws_name, name_list, tag_list):
+    run_list, actual_ws = expand_run_list(store, mongo, ws_name, name_list)
 
     if run_list:
         filter_dict = {"run_name": {"$in": run_list}}
         fields_dict = {"tags": 1}
 
-        records = mongo.get_info_for_runs(workspace, filter_dict, fields_dict)
+        records = mongo.get_info_for_runs(ws_name, filter_dict, fields_dict)
         for record in records:
             run = record["_id"]
             console.print("{}:".format(run))
@@ -156,7 +166,7 @@ def get_run_property_dicts():
         "node": "node_index", "outer": "is_outer", "parent": "is_parent", "path": "path", 
         "pool": "pool", "repeat": "repeat", "restarts": "restarts", "run": "run_name", "run_num": "run_num", 
         "script": "script", "search": "search_type", "search_style": "search_style", "service_type": "service_type", "sku": "sku",
-        "status": "status", "target": "compute", "username": "username", "vc": "vc", "workspace": "ws", "xt_build": "xt_build", 
+        "status": "status", "target": "compute", "username": "username", "vc": "vc", "workspace": "ws_name", "xt_build": "xt_build", 
         "xt_version": "xt_version",
         "hparams": "hparams", "metrics": "metrics", "tags": "tags",
 
@@ -217,9 +227,13 @@ def get_run_property_dicts():
         
 def build_filter_part(fd, args, arg_name, store_name):
     value = args[arg_name]
-    if value:
+
+    if value is not None:
         if isinstance(value, list):
-            fd[store_name] = {"$in": value}
+            if len(value) == 1:
+                fd[store_name] = value[0]
+            else:
+                fd[store_name] = {"$in": value}
         else:
             fd[store_name] = value
 
@@ -227,13 +241,15 @@ def build_run_filter_dict(run_list, user_to_actual, builder, args):
     fd = {}
     option_filters = ["job", "experiment", "target", "service_type", "box", "status", "parent", "child", "outer", "username"]  
 
+    explict_options = qfe.get_explicit_options()
+
     if run_list:
         # filter by specified job names
         fd["_id"] = {"$in": run_list}
 
     # filter by specified options
     for name in option_filters:
-        if name in args:
+        if name in explict_options:
             build_filter_part(fd, args, name, user_to_actual[name])
 
     # filter by filter_list
@@ -277,8 +293,10 @@ def extract_dotted_cols(records, prefix):
 
     return list(nd.keys())
 
-def get_filtered_sorted_limit_runs(store, config, show_gathering, col_dict=None, args=None):
-    
+def get_filtered_sorted_limit_runs(store, config, show_gathering, col_dict=None, args=None, preserve_order=False):
+    # JIT import (workaround cyclic imports)
+    from xtlib.report_builder import ReportBuilder   
+
     console.diag("start of: get_filtered_sorted_limit_runs")
     # required
     run_list = args["run_list"]
@@ -289,13 +307,10 @@ def get_filtered_sorted_limit_runs(store, config, show_gathering, col_dict=None,
     workspace = utils.safe_value(args, "workspace")
     
     if workspace:
-        store.ensure_workspace_exists(workspace, flag_as_error=True)
+        store.ensure_workspace_exists(workspace, flag_as_error=False)
 
     mongo = store.get_mongo()
 
-    # have MONGO update any old RUN documents to new format
-    fixup_mongo_runs.fixup_runs_if_needed(mongo.mongo_db, workspace)
-
     # get info about run properties
     user_to_actual, std_cols_desc = get_run_property_dicts()        
     actual_to_user = {value: key for key, value in user_to_actual.items()}
@@ -304,6 +319,8 @@ def get_filtered_sorted_limit_runs(store, config, show_gathering, col_dict=None,
 
     # get list of specified runs
     pure_run_list, actual_ws = expand_run_list(store, mongo, workspace, run_list)
+    if run_list and not pure_run_list:
+        errors.general_error("no matching runs found")
 
     # build a filter dict for all specified filters
     filter_dict = build_run_filter_dict(pure_run_list, user_to_actual, builder, args)
@@ -315,10 +332,26 @@ def get_filtered_sorted_limit_runs(store, config, show_gathering, col_dict=None,
     records, using_default_last, last = builder.get_mongo_records(mongo, filter_dict, workspace, "runs", actual_to_user, 
         col_dict=col_dict, args=args)
 
+    if preserve_order:
+        records = order_runs_by_user_list(records, pure_run_list)
+
     console.diag("end of: get_filtered_sorted_limit_runs")
 
     return records, using_default_last, user_to_actual, available, builder, last, std_cols_desc
 
+def order_runs_by_user_list(runs, user_list):
+    new_records = list(user_list)
+
+    for run in runs:
+        run_name = run["_id"]
+        index = new_records.index(run_name)
+        if index > -1:
+            new_records[index] = run
+
+    # finally, remove dummy entries
+    new_records = [nr for nr in new_records if isinstance(nr, dict)]
+    return new_records
+
 def list_runs(store, config, args):
 
     records, using_default_last, user_to_actual, available, builder, last, std_cols_desc = \
@@ -383,15 +416,21 @@ def get_run_record(store, workspace, run_name, fields_dict = None):
     rr = run_records[0]
     return rr
 
-def get_service_node_info(store, workspace, run_name):
-    rr = run_helper.get_run_record(store, workspace, name, {"job_id": 1, "node_index": 1})
+def get_job_node_index(store, core, workspace, run_name):
+    rr = get_run_record(store, workspace, run_name, {"job_id": 1, "node_index": 1})
+    job_id = rr["job_id"]
+    node_index = rr["node_index"]
+    return job_id, node_index
+
+def get_service_node_info_with_backend(store, core, workspace, run_name):
+    rr = get_run_record(store, workspace, run_name, {"job_id": 1, "node_index": 1})
     job_id = rr["job_id"]
     node_index = rr["node_index"]
 
-    service_node_info, backend = job_helper.get_service_node_info(job_id, node_index)
-
+    return job_helper.get_service_node_info_with_backend(store, core, workspace, job_id, node_index)
+    
 def get_rightmost_run_num(run):
-    if not run.startswith("run"):
+    if not is_run_name(run):
         errors.syntax_error("Illegal run name, must start with 'run'")
 
     if "." in run:
@@ -436,7 +475,7 @@ def parse_run_list(store, workspace, runs, validate=True):
                 if actual_ws and actual_ws != ws:
                     errors.syntax_error("Cannot mix run_names from different workspaces for this command")
 
-            if not run.startswith("run"):
+            if not is_run_name(run):
                 errors.argument_error("run name", run)
 
             if "-" in run:
@@ -481,13 +520,6 @@ def full_run_name(store_type, ws, run_name):
         full_name = "{}/{}".format(ws, run_name)
     return full_name
 
-def is_simple_run_name(text):
-    is_simple = isinstance(text, str) and text.startswith("run") and len(text) > 3 and text[3].isdigit()
-    if not is_simple:
-        # check for aml run_name:  exper.run34
-        is_simple = isinstance(text, str) and ".run" in text
-    return is_simple
-
 def is_well_formed_run_name(text):
     well_formed = True
     if not "*" in text:
@@ -496,9 +528,9 @@ def is_well_formed_run_name(text):
             parts = text.split("/")
             if len(parts) != 2:
                 well_formed = False
-            elif not is_simple_run_name(parts[1]):
+            elif not is_run_name(parts[1]):
                 well_formed = False
-        elif not is_simple_run_name(text):
+        elif not is_run_name(text):
             well_formed = False
     return well_formed
 
@@ -519,7 +551,7 @@ def validate_run_name(store, ws, run_name, error_if_invalid=True, parse_only=Fal
                 return None, None, None
     return ws, run_name, ws + "/" + run_name
 
-def build_metrics_sets(records, steps=None, merge=False, metrics=None):
+def build_metrics_sets(records, steps=None, merge=False, metrics=None, timebase=None):
     '''
     builds a collection of metrics sets.  A metric set is a set of metrics that have been reported together 
     in the run log.  Here 'together' means as a single log entry.
@@ -540,6 +572,18 @@ def build_metrics_sets(records, steps=None, merge=False, metrics=None):
     last_record = {}
     last_step = None
     merged_records = []
+    need_run_start = False
+    need_first_metric = False
+    time_offset = None
+
+    if timebase == "none":
+        timebase = None
+    elif timebase == "run":
+        need_run_start = True
+    elif timebase == "metric":
+        need_first_metric = True
+    else:
+        errors.UserError("unrecognized timebase value: {}".format(timebase))
 
     if steps:
         step_index = 0
@@ -549,9 +593,19 @@ def build_metrics_sets(records, steps=None, merge=False, metrics=None):
         if not log_dict:
             continue
 
+        if need_run_start and "event" in log_dict and log_dict["event"] == "started":
+            arrow_str = log_dict["time"]
+            time_offset = utils.parse_time(arrow_str).timestamp()
+            need_run_start = False
+
         if not "event" in log_dict or not "data" in log_dict or log_dict["event"] != "metrics":
             continue
 
+        if need_first_metric:
+            arrow_str = log_dict["time"]
+            time_offset = utils.parse_time(arrow_str).timestamp()
+            need_first_metric = False
+
         dd = log_dict["data"]
         
         if step_name is None:
@@ -579,7 +633,11 @@ def build_metrics_sets(records, steps=None, merge=False, metrics=None):
                 continue
 
         # add time to record
-        dd[constants.TIME] = log_dict["time"]
+        arrow_str = log_dict["time"]
+        dt = utils.parse_time(arrow_str)
+        if time_offset:
+            dt = dt - timedelta(seconds=time_offset)
+        dd[constants.TIME] = dt
 
         if metrics:
             # only collect step_name and metrics in "metrics"
@@ -661,16 +719,16 @@ def get_client_cs(core, ws, run_name):
 
     return cs, box_secret
 
-def wrapup_runs_by_node(store, job_id, node_id, node_cancel_result):
+def wrapup_runs_by_node(store, workspace, job_id, node_id, node_cancel_result):
     '''
     wrap up a run from azure batch.  run may have started, or may have completed.  
     '''
     # get job_record
-    job_info = job_helper.get_job_record(store, job_id)
+    job_info = job_helper.get_job_record(store, workspace, job_id)
     ws_name = job_info["ws_name"]
 
     # loads the controller's MRC for context of this node
-    text = store.read_job_file(job_id, constants.FN_MULTI_RUN_CONTEXT)
+    text = store.read_job_file(workspace, job_id, constants.FN_MULTI_RUN_CONTEXT)
     mrc_data = json.loads(text)
 
     context_by_nodes = mrc_data["context_by_nodes"]
@@ -709,4 +767,59 @@ def wrapup_run_with_context(store, run, context_dict):
     store.wrapup_run(context.ws, run_name, context.aggregate_dest, context.dest_name, 
         status, exit_code, context.primary_metric, context.maximize_metric, context.report_rollup, 
         rundir, context.after_files_list, log, capture, job_id=context.job_id, node_id=node_id,
-        run_index=run_index)    
\ No newline at end of file
+        run_index=run_index)    
+
+def get_parent_run_number(name):
+    num = 0
+    if name and name.startswith("run"):
+        part = name[3:]
+
+        if constants.NODE_PREFIX in part:
+            parent, node = part.split(constants.NODE_PREFIX, 1)
+            num = int(parent)
+        elif "." in part:
+            parent, child = part.split(".", 1)
+            num = int(parent)
+        else:
+            num = int(part)
+
+    return num
+
+def get_run_number(name):
+    parent = 0
+    node = 0
+    child = 0
+
+    if name and name.startswith("run"):
+        part = name[3:]
+
+        if constants.NODE_PREFIX in part:
+            parent, node = part.split(constants.NODE_PREFIX, 1)
+            parent = int(parent)
+            node = int(node)
+        elif "." in part:
+            parent, child = part.split(".", 1)
+            parent = int(parent)
+            child = int(child)
+        else:
+            parent = int(part)
+
+        # allow for 1M children, 1M nodes
+        oneM = 1000*1000
+        run_num = oneM*oneM*int(parent) + oneM*int(node) + int(child)
+        
+        return run_num
+
+def is_run_name(name):
+    is_valid = False
+    low_name = name.lower()
+    if low_name.startswith("run"):
+        part = name[3:]
+        if "_" in part:
+            is_valid = part.replace("_", "").isdigit()
+        elif "." in part:
+            is_valid = part.replace(".", "").isdigit()
+        else:
+            is_valid = part.isdigit()
+
+    return is_valid
\ No newline at end of file
