@@ -16,9 +16,10 @@ import logging
 from xtlib import utils
 from xtlib import errors
 from xtlib import pc_utils
-from xtlib.storage import mongo_db
+from xtlib.storage import mongo_v2
 from xtlib import constants
 from xtlib import file_utils
+from xtlib import run_helper
 
 from xtlib.console import console
 from xtlib.helpers.bag import Bag
@@ -35,20 +36,59 @@ from .store_objects import StoreBlobObjs
 
 logger = logging.getLogger(__name__)
 
+# public function
+def create_from_props_dict(pd):
+    return Store(**pd) 
+
 # public function
 def store_from_context(context):
     store = Store(context.store_creds, provider_code_path=context.store_code_path, mongo_conn_str=context.mongo_conn_str)
 
     # set switches for mongo stats, logging, and scheduling
     mongo = store.mongo
-    mongo.update_job_stats = context.update_job_stats
-    mongo.update_run_stats = context.update_run_stats
-    mongo.add_log_records = context.add_log_records
+
+    if mongo:
+        mongo.update_job_stats = context.update_job_stats
+        mongo.update_run_stats = context.update_run_stats
+        mongo.add_log_records = context.add_log_records
+
+    return store
+
+# public function
+def Store(storage_creds=None, config=None, max_retries=10, run_cache_dir=None, feedback_enabled=True,  mongo_conn_str=None, 
+        provider_code_path=None, validate_creds=False):
+    '''
+    This function peeks at the version number in the storage and chooses StoreV1 or StoreV2 as the store 
+    class to return an instance of.
+    '''
+    if config:
+        if not storage_creds:
+            storage_creds = config.get_storage_creds()
+
+        if not provider_code_path:
+            provider_code_path = config.get_storage_provider_code_path(storage_creds)
+
+    # instantiate the provider
+    class_ctr = utils.get_class_ctr(provider_code_path)
+    provider = class_ctr(storage_creds)
+
+    # does the version 1 system info container exist?
+    use_v1 = provider.does_container_exist(constants.INFO_CONTAINER_V1)
+
+    if use_v1:
+        # JIT import to minimize impact of V1
+        from xtlib.storage_v1.store_v1 import StoreV1
+
+        store = StoreV1(storage_creds, config, max_retries, run_cache_dir, feedback_enabled,  mongo_conn_str, 
+            provider_code_path, validate_creds)
+    else:
+        store = StoreV2(storage_creds, config, max_retries, run_cache_dir, feedback_enabled,  mongo_conn_str, 
+            provider_code_path, validate_creds)
 
     return store
 
 # main class that implements the XT STORE API
-class Store():
+class StoreV2():
     '''This class provides access to the XT Store, which is based on a storage provider.
     Methods are provided to manage workspaces, experiments, runs, and related files.
 
@@ -74,6 +114,10 @@ class Store():
         provider_code_path=None, validate_creds=False):
         '''This is the constructor for the Store class. '''
 
+        # # debug issue of using correct storage account on compute node
+        # if os.getenv("XT_STORE_CREDS"):
+        #     utils.debug_break()
+
         self.feedback_enabled = feedback_enabled
 
         if not mongo_conn_str and config:
@@ -100,7 +144,7 @@ class Store():
                 errors.service_error("Azure Storage service credentials not set correctly" + 
                     "; use 'xt config' to correct")
 
-        self.mongo = mongo_db.MongoDB(mongo_conn_str, run_cache_dir) if mongo_conn_str else None
+        self.mongo = mongo_v2.MongoDB2(mongo_conn_str, run_cache_dir=run_cache_dir) if mongo_conn_str else None
         self.mongo_conn_str = mongo_conn_str
         self.run_cache_dir = run_cache_dir
         self.provider_code_path = provider_code_path
@@ -111,8 +155,12 @@ class Store():
         self.cap_stdout = None
         self.cap_stderr = None
 
-        self.helper.validate_storage_and_mongo(self.mongo)
+        if self.mongo:
+            self.helper.validate_storage_and_mongo(self.mongo)
 
+    def get_name(self):
+        return self.helper.get_name()
+        
     def get_mongo(self):
         return self.mongo
 
@@ -121,10 +169,6 @@ class Store():
             "run_cache_dir": self.run_cache_dir, "provider_code_path": self.provider_code_path}
         return pd
 
-    @staticmethod
-    def create_from_props_dict(pd):
-        return Store(**pd) 
-
     def _error(self, msg):
         raise Exception("Error - {}".format(msg))
 
@@ -161,7 +205,9 @@ class Store():
     # ---- WORKSPACE ----
 
     def ensure_workspace_exists(self, ws_name, flag_as_error=True):
-        return self.helper.ensure_workspace_exists(ws_name, flag_as_error)
+        created = self.helper.ensure_workspace_exists(ws_name, flag_as_error)
+        if True:  # created:
+            self.mongo.create_workspace_if_needed(ws_name)
 
     def get_running_workspace(self):
         ''' returns the name of the workspace associated with the current XT run.
@@ -173,21 +219,24 @@ class Store():
         '''
         return self.helper.does_workspace_exist(ws_name)
 
-    def create_workspace(self, ws_name, description=None):
+    def create_workspace(self, ws_name, database=None):
         ''' create a new workspace using the specified name.
         '''
-        self.helper.create_workspace(ws_name, description)
+        self.helper.create_workspace(ws_name)
 
         # log some information
-        self.log_workspace_event(ws_name, "created", {"description": description})
+        self.log_workspace_event(ws_name, "created", {"database": database})
+
+        self.mongo.create_workspace_if_needed(ws_name, database)
 
     def delete_workspace(self, ws_name):
         ''' delete the specified workspace, and all of the runs stored within it.
         '''
+
         result = self.helper.delete_workspace(ws_name)
         if result:
             # remove associated mongo info
-            self.mongo.remove_workspace(ws_name)
+            self.mongo.delete_workspace_if_needed(ws_name)
 
         return result
     
@@ -200,10 +249,17 @@ class Store():
         # append to workspace log file
         self.append_workspace_file(ws_name, WORKSPACE_LOG, rd_text + "\n")
 
-    def get_workspace_names(self):
+    def get_workspace_names(self, use_mongo=False):
         ''' return the names of all workspaces that are currently defined in the XT Store.
         '''
-        return self.helper.get_workspace_names()
+        if use_mongo:
+            # use mongo for more faster workspace names 
+            names = self.mongo.get_workspace_names()
+        else:
+            # storage is the ground truth for workspace names
+            names = self.helper.get_workspace_names()
+
+        return names
 
     def is_legal_workspace_name(self, name):
         ''' return True if 'name' is a legal workspace name for the current XT Store.
@@ -282,8 +338,8 @@ class Store():
         wf = self.workspace_files(ws_name, use_blobs=True)
         return wf.does_file_exist(ws_fn)
 
-    def get_job_workspace(self, job_id):
-        return self.mongo.get_job_workspace(job_id)
+    # def get_job_workspace(self, job_id):
+    #     return self.mongo.get_job_workspace(job_id)
 
     # ---- EXPERIMENT ----
 
@@ -423,12 +479,19 @@ class Store():
         if not default_next:
             default_next = 1
 
-        run_id = self.mongo.get_next_run_id(ws_name, default_next=default_next)
-        run_name = "run{}".format(run_id)
+        # allow for pending runs to write to our recycled workspace (deleted and then recreated)
+        # this is a common problem with the current quicktest and the "xt-demo" workspace
+        while True:
+            run_id = self.mongo.get_next_run_id(ws_name, default_next=default_next)
+            run_name = "run{}".format(run_id)
 
-        # ensure we are not somehow overwriting an existing run
-        exists = self.does_run_exist(ws_name, run_name)
-        assert not exists
+            # ensure we are not somehow overwriting an existing run
+            exists = self.does_run_exist(ws_name, run_name)
+            if exists:
+                errors.warning("skipping allocation of: run_name={}, workspace={} (found existing run in storage)".format( \
+                    run_name, ws_name))
+            else:
+                break
 
         return self.helper.create_next_run_by_name(ws_name, run_name)
 
@@ -467,18 +530,6 @@ class Store():
         #console.print("returning hparams=", hparams)
         return hparams
 
-    def get_run_num(self, run_name):
-        base = run_name[3:]
-
-        if "." in base:
-            parent, child = base.split(".")
-            # allow for 1 million 
-            run_num = 1000*1000*int(parent) + int(child)
-        else:
-            run_num = 1000*1000*int(base)
-
-        return run_num
-
     def start_child_run(self, ws_name, parent_run_name, exper_name=None, cmd_line_args=None, all_args=None, description=None, 
             from_ip=None, from_host=None, app_name=None, repeat=None, box_name="local", job_id=None, pool=None, 
             node_index=None, username=None, aggregate_dest=None, path=None, is_aml=False, child_name=None, compute=None, 
@@ -492,7 +543,7 @@ class Store():
             # child_name = parent_run_name + "_" + str(child_num)
             child_name = aml_child_name
         else:
-            child_name = self._create_next_child_directory(ws_name, parent_run_name, child_name=child_name)
+            self._create_next_child_directory(ws_name, parent_run_name, child_name=child_name)
 
         # we are called from controller, so our box_name is the name of this machine
         hostname = pc_utils.get_hostname()
@@ -501,71 +552,18 @@ class Store():
         if box_name == "local":
             box_name = hostname
 
-        # for cases where workspaces are deleted, renamed, etc, this gives a truely unique id for the run
-        run_guid = str(uuid.uuid4())
+        self.start_run_core(ws_name, child_name, exper_name=exper_name, description=description, username=username,
+            box_name=box_name, app_name=app_name, repeat=None, is_parent=False, job_id=job_id, pool=pool, node_index=node_index,
+            aggregate_dest=aggregate_dest, path=path, aml_run_id=None, compute=compute, service_type=service_type, search_type=search_type,
+            sku=sku, search_style=search_style, is_child=True)           
 
-        run_num = self.get_run_num(child_name)
-        script = os.path.basename(path) if path else None
-        create_time = utils.get_time()
-
-        build_parts = constants.BUILD.split(",")
-        xt_version = build_parts[0].split(":")[1].strip()
-        xt_build = build_parts[1].split(":")[1].strip()
-
-        dd = {"ws": ws_name, "run_name": child_name, "run_num": run_num, "run_guid": run_guid, "description": description, "exper_name": exper_name, 
-            "job_id": job_id, "node_index": node_index, "is_outer": False, "is_parent": False, "is_child": True,
-            "from_ip": from_ip, "from_computer_name": from_host, "username": username, 
-            "box_name": box_name, "app_name": app_name, "repeat": None, "path": path, "script": script, "create_time": create_time,
-            "compute": compute, "service_type": service_type, "sku": sku, "search_style": search_style,
-            "search_type": search_type,  "xt_build": xt_build, "xt_version": xt_version, "run_index": run_index}
-
-        self.mongo.create_mongo_run(dd)
-
-        # log "created" event for child run
-        self.log_run_event(ws_name, child_name, "created", dd, is_aml=is_aml)
-
-        # append "start" record to workspace summary log
-        dd["time"] = utils.get_time()
-        dd["event"] = "created"
-        text = json.dumps(dd) + "\n"    
-
-        if not is_aml:
-            # append to summary file in run dir
-            self.append_run_file(ws_name, child_name, constants.RUN_SUMMARY_LOG, text)
-
-        if cmd_line_args:
-            self.log_run_event(ws_name, child_name, "cmd_line_args", {"data": cmd_line_args}, is_aml=is_aml)  
-        if all_args:
-            self.log_run_event(ws_name, child_name, "all_args", {"data": all_args}, is_aml=is_aml)
-    
         # finally, add a "child_created" record to the parent
         self.log_run_event(ws_name, parent_run_name, "child_created", {"child_name": child_name}, is_aml=is_aml)
-        
-        return child_name
-
-    def start_run(self, ws_name, exper_name=None, description=None, username=None,
-            box_name=None, app_name=None, repeat=None, is_parent=False, job_id=None, pool=None, node_index=None,
-            aggregate_dest="none", path=None, compute=None, service_type=None, search_type=None, sku=None,
-            search_style=None):
-        '''
-        This is usually called from the XT client machine (where user is running XT), so we can get from_ip and
-        from_computer_name from the local machine.
-        '''
-
-        console.diag("start_run: start")
-
-        # helper uses a lock to ensure atomic operation here 
-        run_name = self._create_next_run_directory(ws_name, is_parent)
-
-        return self.start_run_core(ws_name, run_name=run_name, exper_name=exper_name, description=description, username=username,
-            box_name=box_name, app_name=app_name, repeat=repeat, is_parent=is_parent, job_id=job_id, 
-            node_index=node_index, aggregate_dest=aggregate_dest, path=path, compute=compute, service_type=service_type, 
-            search_type=search_type, sku=sku, search_style=search_style)
 
     def start_run_core(self, ws_name, run_name, exper_name=None, description=None, username=None,
             box_name=None, app_name=None, repeat=None, is_parent=False, job_id=None, pool=None, node_index=None,
             aggregate_dest="none", path=None, aml_run_id=None, compute=None, service_type=None, search_type=None,
-            sku=None, search_style=None):
+            sku=None, search_style=None, is_child=False):
 
         console.diag("start_run: after create_next_run_directory")
 
@@ -585,7 +583,7 @@ class Store():
 
         # for cases where workspaces are deleted, renamed, etc, this gives a truely unique id for the run
         run_guid = str(uuid.uuid4())
-        run_num = self.get_run_num(run_name)
+        run_num = run_helper.get_run_number(run_name)
         script = os.path.basename(path) if path else None
         create_time = utils.get_time()
 
@@ -593,18 +591,41 @@ class Store():
         xt_version = build_parts[0].split(":")[1].strip()
         xt_build = build_parts[1].split(":")[1].strip()
 
-        dd = {"ws": ws_name, "run_name": run_name, "run_num": run_num, "run_guid": run_guid, "description": description, "exper_name": exper_name, 
-            "job_id": job_id, "node_index": node_index, "is_outer": True, "is_parent": is_parent, "is_child": False, 
-            "from_ip": ip, "from_computer_name": hostname, "username": username, 
-            "box_name": box_name, "app_name": app_name, "repeat": repeat, "path": path, "script": script, "create_time": create_time,
-            "node": node_index, "compute": compute, "service_type": service_type, "search_type": search_type, "sku": sku,
-            "xt_build": xt_build, "xt_version": xt_version, "search_style": search_style, "run_index": None}
+        dd = {
+            # RUN_INFO: flat READONLY props (27)
+            "app_name": app_name, 
+            "box_name": box_name, 
+            "compute": compute, 
+            "create_time": create_time,
+            "description": description, 
+            "exper_name": exper_name, 
+            "from_ip": ip, 
+            "from_computer_name": hostname, 
+            "is_child": is_child, 
+            "is_outer": True, 
+            "is_parent": is_parent, 
+            "job_id": job_id, 
+            "node_index": node_index, 
+            "path": path, 
+            "repeat": repeat, 
+            "script": script, 
+            "run_name": run_name, 
+            "run_num": run_num, 
+            "run_guid": run_guid, 
+            "search_style": search_style, 
+            "search_type": search_type, 
+            "service_type": service_type, 
+            "sku": sku,
+            "username": username, 
+            "ws_name": ws_name, 
+            "xt_build": xt_build, 
+            "xt_version": xt_version, 
+            }
 
         if aml_run_id:
             dd["aml_run_id"] = aml_run_id
 
-        self.mongo.create_mongo_run(dd)
-
+        # this call will also create run in mongo, as appropriate
         self.log_run_event(ws_name, run_name, "created", dd, is_aml=is_aml)
 
         console.diag("start_run: after log_run_event")
@@ -636,7 +657,8 @@ class Store():
             end_time = utils.get_time()
 
         self.log_run_event(ws_name, run_name, "ended", {"status": status, "exit_code": exit_code, \
-            "metrics_rollup": metrics_rollup_dict}, event_time=end_time, is_aml=is_aml)
+            "restarts": restarts, "metrics_rollup": metrics_rollup_dict}, \
+                event_time=end_time, is_aml=is_aml)
 
         if not is_aml:
             # append "end" record to workspace summary log
@@ -688,7 +710,7 @@ class Store():
                     if aggregate_dest == "experiment":
                         self.append_experiment_file(ws_name, dest_name, constants.ALL_RUNS_FN, text)
                     elif aggregate_dest == "job":
-                        self.append_job_file(dest_name, constants.ALL_RUNS_FN, text)
+                        self.append_job_file(ws_name, dest_name, constants.ALL_RUNS_FN, text)
 
             # LOG END RUN
             log_records = self.get_run_log(ws_name, run_name)
@@ -700,7 +722,8 @@ class Store():
         self.end_run(ws_name, run_name, status, exit_code, hparams, metrics, restarts=restarts, 
             end_time=end_time, aggregate_dest=aggregate_dest, dest_name=dest_name, is_aml=is_aml)
 
-        self.mongo.update_mongo_run_at_end(ws_name, run_name, status, exit_code, restarts, end_time, log_records, hparams, metrics)
+        if self.mongo:
+            self.mongo.update_mongo_run_at_end(ws_name, run_name, status, exit_code, restarts, end_time, log_records, hparams, metrics)
 
     def _roll_up_hparams(self, log_records):
         hparams_dict = {}
@@ -785,6 +808,7 @@ class Store():
 
         if data_dict and not isinstance(data_dict, dict):   
             raise Exception("data_dict argument is not a dict: " + str(data_dict))
+
         record_dict = {"time": event_time, "event": event_name, "data": data_dict}
         #console.print("record_dict=", record_dict)
 
@@ -793,14 +817,15 @@ class Store():
             # append to run log file
             self.append_run_file(ws_name, run_name, RUN_LOG, rd_text + "\n")
 
-        # log all backend types to mongo
-        self.mongo.add_run_event(ws_name, run_name, record_dict)
+        if self.mongo:
+            # log all backend types to mongo
+            self.mongo.process_run_event(ws_name, run_name, event_name, data_dict, record_dict)
 
     def get_job_names(self, filter_dict=None, fields_dict=None):
         return self.mongo.get_job_names(filter_dict, fields_dict)
 
-    def get_ws_runs(self, ws_name, filter_dict=None, include_log_records=False, first_count=None, last_count=None, sort_dict=None):
-        return self.mongo.get_ws_runs(ws_name, filter_dict, include_log_records, first_count, last_count, sort_dict)
+    def get_job_names_from_storage(self):
+        return self.helper.get_job_names()
 
     def get_all_runs(self, aggregator_dest, ws_name, job_or_exper_name, filter_dict=None, fields_dict=None, use_cache=False, 
         fn_cache=None, first_count=None, last_count=None, sort_dict=None, batch_size=None):
@@ -829,7 +854,7 @@ class Store():
                 from_path = os.path.dirname(output_files)
                 to_path = "after/" + os.path.basename(from_path) if from_path else "after" 
                 output_files = os.path.abspath(file_utils.path_join(rundir, output_files))
-                console.print("\nprocessing AFTER: ws_name=", ws_name, ", run_name=", run_name, ", output_files=", output_files, ", from_path=", from_path, ", to_path=", to_path)
+                #console.print("\nprocessing AFTER: ws_name=", ws_name, ", run_name=", run_name, ", output_files=", output_files, ", from_path=", from_path, ", to_path=", to_path)
 
                 copied = self.upload_files_to_run(ws_name, run_name, to_path, output_files, exclude_dirs_and_files=after_omit_list)
                 #console.print("AFTER files copied=", copied)
@@ -842,18 +867,25 @@ class Store():
 
         # tell mongo RUNS that this run has completed
         console.diag("calling MONGO run_exit: ws={}, run_name={}".format(ws_name, run_name))
-        self.mongo.run_exit(ws_name, run_name)
 
-        if not is_parent:
-            # tell mongo this job has a completed run
-            console.diag("calling MONGO job_run_exit: job_id={}".format(job_id))
-            self.mongo.job_run_exit(job_id, exit_code)
+        if self.mongo:
+            self.mongo.run_exit(ws_name, run_name)
+
+            if not is_parent:
+                # tell mongo this job has a completed run
+                console.diag("calling MONGO job_run_exit: job_id={}".format(job_id))
+                self.mongo.job_run_exit(ws_name, job_id, exit_code)
 
     def copy_run_files_to_run(self, ws_name, from_run, run_wildcard, to_run, to_path):
         return self.helper.copy_run_files_to_run(ws_name, from_run, run_wildcard, to_run, to_path)
 
     def update_mongo_run_at_end(self, ws_name, run_name, status, exit_code, restarts, end_time, log_records, hparams, metrics):
-        return self.mongo.update_mongo_run_at_end(ws_name, run_name, status, exit_code, restarts, end_time, log_records, hparams, metrics)
+        result = None
+
+        if self.mongo:
+            result = self.mongo.update_mongo_run_at_end(ws_name, run_name, status, exit_code, restarts, end_time, log_records, hparams, metrics)
+
+        return result
 
     #---- LOW LEVEL ----
             
@@ -934,129 +966,130 @@ class Store():
 
     #---- JOBS ----
 
-    def create_job(self):
-        # read legacy next_job_info in case we are first to run w/mongo-sequence numbers
-        default_next = self.helper.read_legacy_next_job_id()
-        if not default_next:
-            default_next = 1
-            
-        job_id = self.mongo.get_next_job_id(default_next=default_next)
+    def create_job(self, ws_name):
+        job_num = self.mongo.get_next_job_id(ws_name, "next_job_number")
+
+        job_name = "job{}".format(job_num)
+
+        # create LAST JOB CREATED blob 
+        blob_fn = constants.WORKSPACE_DIR + "/" + constants.WORKSPACE_LAST_JOB
+        self.helper._create_blob(ws_name, blob_fn, str(job_num))
 
-        job_name = "job{}".format(job_id)
         return job_name
 
-    def get_job_secret(self, job_id):
-        records = self.mongo.get_info_for_jobs( filter_dict={"_id": job_id}, fields_dict={"job_secret": 1} )
+    def get_job_secret(self, ws_name, job_id):
+        key = "{}/{}".format(ws_name, job_id)
+
+        records = self.mongo.get_info_for_jobs( filter_dict={"_id": key}, fields_dict={"job_secret": 1} )
         value = records[0]["job_secret"] if len(records) else None
         return value
 
-    def read_job_info_file(self, job_id):
-        return self.helper.read_job_info_file(job_id)
+    def read_job_info_file(self, ws_name, job_id):
+        return self.helper.read_job_info_file(ws_name, job_id)
 
-    def write_job_info_file(self, job_id, text):
-        # write to XT store
-        self.helper.write_job_info_file(job_id, text)
+    def log_job_info(self, ws_name, job_id, dd):
+        text = json.dumps(dd, indent=4)
+        self.helper.write_job_info_file(ws_name, job_id, text)
 
         # update mongo-db for fast access
-        dd = json.loads(text)
-        self.mongo.update_job_info(job_id, dd)
-    
-    def log_job_info(self, job_id, dd):
-        text = json.dumps(dd, indent=4)
-        self.write_job_info_file(job_id, text)
+        self.mongo.update_job(dd)
 
-    def log_job_event(self, job_id, event_name, data_dict=None, description=None, event_time=None):
+    def log_job_event(self, ws_name, job_id, event_name, data_dict=None, event_time=None):
         if not event_time:
             event_time = utils.get_time()
 
         if data_dict and not isinstance(data_dict, dict):   
             raise Exception("data_dict argument is not a dict: " + str(data_dict))
-        record_dict = {"time": event_time, "event": event_name, "data": data_dict, "description": description}
+        record_dict = {"time": event_time, "event": event_name, "data": data_dict}
         #console.print("record_dict=", record_dict)
 
         rd_text = json.dumps(record_dict)
         # append to run log file
-        self.append_job_file(job_id, constants.JOB_LOG, rd_text + "\n")
+        self.append_job_file(ws_name, job_id, constants.JOB_LOG, rd_text + "\n")
 
-    def get_job_log(self, job_id):
-        text = self.read_job_file(job_id, constants.JOB_LOG)
+    def get_job_log(self, ws_name, job_id):
+        text = self.read_job_file(ws_name, job_id, constants.JOB_LOG)
         records = utils.load_json_records(text)
         return records
 
     #---- JOB FILES ----
 
-    def create_job_file(self, job_name, job_path, text):
+    def create_job_file(self, ws_name, job_name, job_path, text):
         '''create a job file specified by 'job_path' from the text 'text'.
         '''
         #self.helper.create_job_file(job_name, job_path, text)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.create_file(job_path, text)
 
-    def append_job_file(self, job_name, job_path, text):
+    def append_job_file(self, ws_name, job_name, job_path, text):
         '''append text 'text' to the job file 'job_path'.
         '''
         #self.helper.append_job_file(job_name, job_path, text)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.append_file(job_path, text)
 
-    def read_job_file(self, job_name, job_path):
+    def read_job_file(self, ws_name, job_name, job_path):
         '''return the contexts of the job file 'job_path'.
         '''
         #return self.helper.read_job_file(job_name, job_path)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.read_file(job_path)
 
-    def upload_file_to_job(self, job_name, fn_job, fn_source):
+    def upload_file_to_job(self, ws_name, job_name, fn_job, fn_source):
         '''upload the local file 'fn_source' as the job file 'fn_job'.
         '''
         #return self.helper.upload_file_to_job(job_name, job_folder, fn_source)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.upload_file(fn_job, fn_source)
 
-    def upload_files_to_job(self, job_name, job_folder, source_wildcard, recursive=False, exclude_dirs_and_files=[]):
+    def upload_files_to_job(self, ws_name, job_name, job_folder, source_wildcard, recursive=False, exclude_dirs_and_files=[]):
         '''upload the local files specified by 'source_wildcard' to the job folder 'job_folder'.
         '''
         #return self.helper.upload_files_to_job(job_name, job_folder, source_wildcard)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.upload_files(job_folder, source_wildcard, recursive=recursive, exclude_dirs_and_files=exclude_dirs_and_files)
 
-    def download_file_from_job(self, job_name, job_fn, dest_fn):
+    def download_file_from_job(self, ws_name, job_name, job_fn, dest_fn):
         '''download the job file 'job_fn' to the local file 'dest_fn'.
         '''
         #return self.helper.download_file_from_job(job_name, job_fn, dest_fn)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.download_file(job_fn, dest_fn)
 
-    def download_files_from_job(self, job_name, job_wildcard, dest_folder):
+    def download_files_from_job(self, ws_name, job_name, job_wildcard, dest_folder):
         #return self.helper.download_files_from_job(job_name, job_wildcard, dest_folder)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.download_files(job_wildcard, dest_folder)
 
-    def get_job_filenames(self, job_name, job_wildcard=None, full_paths=False):
+    def get_job_filenames(self, ws_name, job_name, job_wildcard=None, full_paths=False):
         '''return the names of the job files specified by 'job_wildcard'.
         '''
         #return self.helper.get_run_filenames(ws_name, run_name, run_wildcard)    
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.get_filenames(job_wildcard, full_paths=full_paths)
 
-    def delete_job_file(self, job_name, filename):
+    def delete_job_file(self, ws_name, job_name, filename):
         '''delete the job files specified by 'job_wildcard'.
         '''
         #return self.helper.delete_run_files(ws_name, run_name, run_wildcard)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.delete_file(filename)
 
-    def does_job_file_exist(self, job_name, job_path):
+    def does_job_file_exist(self, ws_name, job_name, job_path):
         '''return True if the job file 'job_path' exists.
         '''
         #return self.helper.does_job_file_exist(job_name, job_path)
-        jf = self.job_files(job_name, use_blobs=True)
+        jf = self.job_files(ws_name, job_name, use_blobs=True)
         return jf.does_file_exist(job_path)
 
+    def does_job_exist(self, ws_name, job_id):
+        return self.helper.does_job_exist(ws_name, job_id)
+
+
     # def append_job_run_name(self, job_name, run_name):
     #     return self.helper.append_job_run_name(job_name, run_name)
 
-    def get_job_run_names(self, job_name):
+    def get_job_run_names(self, ws_name, job_name):
         return self.helper.get_job_run_names(job_name)
 
     # ---- DIRECT ACCESS ----
@@ -1096,10 +1129,20 @@ class Store():
     def experiment_files(self, ws_name, exper_name, use_blobs=False):
         return self.helper.experiment_files(ws_name, exper_name, use_blobs)
 
-    def job_files(self, job_name, use_blobs=False):
-        return self.helper.job_files(job_name, use_blobs)
+    def job_files(self, ws_name, job_name, use_blobs=False):
+        return self.helper.job_files(ws_name, job_name, use_blobs)
+
 
 # sample code for path objects
 #   wp = store.WorkspaceFiles(ws_name="ws1")  
 #   wp.create_file("test.txt", "this is test.txt contents")
 
+#---- share helpers ----
+def make_share_name(name):
+    return "00-share-{}".format(name)
+
+def is_share_name(name):
+    return isinstance(name, str) and name.startswith("00-share-")
+
+
+
